#include <stdio.h>
#include <errno.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/time.h>
#include <fcntl.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <sys/io.h>

#define SERIALDEVICE "/dev/ttyS0"

#define INITTIM	200000
#define WAITTIM	2000
#define ENDTIM	500000

int serfd;
int serio;
char *argv0;
char *xchan[]={
	"a0110", "b0111", "c0100", "d0101", "e1000", "f1001", "g1010", "h1011",
	"i1110", "j1111", "k1100", "l1101", "m0000", "n0001", "o0010", "p0011",
	NULL
};

usage()
{
	fprintf(stderr, "Usage: %s [/dev/ttyS*] [-v] <CMD>...\n", argv0);
	fprintf(stderr, "\tWhere CMD can be:\n");
	fprintf(stderr, "\t\t[a-p][1-16]on\n");
	fprintf(stderr, "\t\t[a-p][1-16]off\n");
	fprintf(stderr, "\t\t[a-p]brite\n");
	fprintf(stderr, "\t\t[a-p]dim\n");
	fprintf(stderr, "\t\tsleep[<N>]> - N = seconds, default 1\n");
	exit(1);
}

main(argc, argv)
int argc;
char **argv;
{
	int argn=1;
	int verb=0;
	int err=0;
	int secs;
	char *cmd, *p;
	char xch, **xchp;
	int xno;
	int xonoff;

	argv0=argv[0];
	if ((p=strrchr(argv0, '/'))) {
		argv0=p;
		argv0++;
	}
	if (argc < 2)
		usage();

	/* set and open device - often must be root */
	if (strncmp(argv[argn], "/dev/", 5) == 0) {
		cmd=argv[argn];
		argn++;
	} else
		cmd=SERIALDEVICE;
	if ((serfd=open(cmd, O_RDWR|O_NDELAY)) == -1) {
		perror(cmd);
		exit(1);
	}
	iopl(3);
	ioctl(serfd, TIOCMGET, &serio);

	for (; argn < argc ; argn++) {
		cmd=argv[argn];
		for (p=cmd; *p; p++)
			if (isupper(*p))
				*p = tolower(*p);

		/* verbose option */
		if (strcmp(cmd, "-v") == 0) {
			verb=1;
			continue;
		}

		/* sleep option */
		if (strncmp(cmd, "sleep", 5) == 0) {
			if (isdigit(cmd[5]))
				secs=atoi(&cmd[5]);
			else
				secs=1;
			if (verb)
				printf("%s: command \"%s\" OK\n", argv0, cmd);
			sleep(secs);
			continue;
		}

		/* format <[a-p]><[1-16]><on|off> or <a-p><brite|dim> */
		xch=cmd[0];
		for (xchp=xchan; *xchp; xchp++)
			if (**xchp == xch)
				break;
		if (*xchp == NULL)
			goto cmderr;
		p = &cmd[1];

		if (strcmp(p, "brite") == 0) {
			x10header();
			p = *xchp;
			x10sends(&p[1]);
			x10sends("0000");
			x10sends("10001000");
			x10footer();
			if (verb)
				printf("%s: command \"%s\" OK\n", argv0, cmd);
			continue;
		}

		if (strcmp(p, "dim") == 0) {
			x10header();
			p = *xchp;
			x10sends(&p[1]);
			x10sends("0000");
			x10sends("10011000");
			x10footer();
			if (verb)
				printf("%s: command \"%s\" OK\n", argv0, cmd);
			continue;
		}

		xno=atoi(p);
		if (xno < 1 || xno > 16)
			goto cmderr;
		xno--;
		while (isdigit(*p))
			p++;
		if (strcmp(p, "on") != 0 && strcmp(p, "off") != 0)
			goto cmderr;
		p++;
		xonoff=((*p == 'f')?1:0);

		x10header();
		p = *xchp;
		x10sends(&p[1]);
		x10sends("0");
		x10sendn(xno & 8);
		x10sends("00 0");
		x10sendn(xno & 4);
		x10sendn(xonoff);
		x10sendn(xno & 1);
		x10sendn(xno & 2);
		x10sends("000");
		x10footer();
		if (verb)
			printf("%s: command \"%s\" OK\n", argv0, cmd);
		continue;

cmderr:
		if (verb)
			printf("%s: command \"%s\" UNKNOWN\n", argv0, cmd);
		else
			fprintf(stderr, "%s: Unknown command \"%s\"\n", argv0, cmd);
		err=1;
	}
	if (err)
		usage();
	exit(0);
}

/* delay for microsconds */
musleep(ms)
unsigned ms;
{
	struct timeval tbeg, tnow;

	/* brute force works more accurately than interrupts */
	gettimeofday(&tbeg, NULL);
	while (1) {
		gettimeofday(&tnow, NULL);
		tnow.tv_sec -= tbeg.tv_sec;
		if (tnow.tv_usec < tbeg.tv_usec) {
			tnow.tv_usec += 1000000;
			tnow.tv_sec  -= 1;
			tnow.tv_usec -= tbeg.tv_usec;
		} else
			tnow.tv_usec -= tbeg.tv_usec;
		if ((tnow.tv_sec*1000000+tnow.tv_usec) >= ms)
			return;
	}
}

x10wait(ms)
unsigned ms;
{
	/* set DTR high and RTS high */
	serio |= TIOCM_DTR;
	serio |= TIOCM_RTS;
	ioctl(serfd, TIOCMSET, &serio);
	musleep(ms);
}

x10reset()
{
	/* set DTR low and RTS low */
	serio &= ~TIOCM_DTR;
	serio &= ~TIOCM_RTS;
	ioctl(serfd, TIOCMSET, &serio);
	musleep(INITTIM);
	x10wait(INITTIM);
}

x10logic0()
{
	/* set DTR high and RTS low */
	serio |= TIOCM_DTR;
	serio &= ~TIOCM_RTS;
	ioctl(serfd, TIOCMSET, &serio);
	musleep(WAITTIM);
	x10wait(WAITTIM);
}

x10logic1()
{
	/* set DTR low and RTS high */
	serio &= ~TIOCM_DTR;
	serio |= TIOCM_RTS;
	ioctl(serfd, TIOCMSET, &serio);
	musleep(WAITTIM);
	x10wait(WAITTIM);
}

x10sends(s)
char *s;
{
	for (;*s;s++) {
		if (*s == '0')
			x10logic0();
		if (*s == '1')
			x10logic1();
	}
}

x10sendn(n)
unsigned n;
{
	if (n)
		x10logic1();
	else
		x10logic0();
}

x10header()
{
	x10reset();
	x10sends("11010101 10101010");
}

x10footer()
{
	x10sends("10101101");
	x10wait(ENDTIM);
}
